import os
import os
import nbformat as nbf
import argparse
import json
import sys
from pathlib import Path
import shutil

proj_folder = os.path.dirname(os.path.abspath(__file__))
sys.path.append(proj_folder)
from properties.propertyFinder import PropertyFinder
from assert_gen.notebookRunner import NotebookRunner
from assert_gen.assertionGenerator import AssertionGenerator
from utils.pytest_run import run_missing_tests, aggregate_results_to_csv
from utils.utils import setup_logger, get_notebook_name

config_path = os.path.join(proj_folder, "config.json")
global CONFIG_FILEPATH

with open(config_path, "r") as config_file:
    CONFIG_FILEPATH = json.load(config_file)["FILE_PATH"]

def main():
    parser = argparse.ArgumentParser(description='Tool for generating tests for Jupyter Notebooks')
    parser.add_argument('notebook', type=str, help='The path to the Jupyter notebook file')
    parser.add_argument('-n', default="30", help='Number of iterations for generating assertions (default: 30)')
    parser.add_argument('-o', required=True, help='Absolute path for output files')
    parser.add_argument("-m", default="chebyshev", choices=["hoeffding", "chebyshev"], help="Method for bound calculation.")
    parser.add_argument('-c', default=0.95, help='Confidence level for the bounds calculation (default: 0.95)')
    parser.add_argument("--pytest", action="store_true", help="Run pytest with nbtest plugin to test the assertions in the notebook")
    parser.add_argument('-pn', default="1", help='Number of iterations for running the notebook when --pytest is provided')
    parser.add_argument('--rundir', default='./', help='Folder to run the notebook. This folder should have the file structure required for the notebook')
    args = parser.parse_args()

    global stats_method, conf_level, output_dir, iterations, notebook_path
    stats_method = args.m
    conf_level = float(args.c)
    output_dir = os.path.abspath(args.o)

    iterations = int(args.n)
    pytest_iterations = int(args.pn)
    notebook_path = os.path.abspath(args.notebook)

    rundir = os.path.abspath(args.rundir)
    original_nb_folder = Path(notebook_path).parent
    os.makedirs(output_dir, exist_ok=True)

    # Paths
    notebook_fname = get_notebook_name(notebook_path)

    tmp_ipynb = os.path.join(original_nb_folder, f"{notebook_fname}_{CONFIG_FILEPATH['TMP_IPYNB']}")
    instrumentation_pkl = os.path.join(output_dir, f"{notebook_fname}_{CONFIG_FILEPATH['INSTRUMENTATION_PKL']}")
    properties_csv = os.path.join(output_dir, f"{notebook_fname}_{CONFIG_FILEPATH['PROPERTIES_CSV']}")

    if args.pytest:
        logs_dir = os.path.join(output_dir, f"{notebook_fname}_logs")
        os.makedirs(logs_dir, exist_ok=True)
        logger = setup_logger(os.path.join(logs_dir, f'main.log'), 'main')
        if os.path.exists(notebook_path):
            # Run pytest over notebooks to test assertions
            run_missing_tests(output_dir, notebook_path, pytest_iterations, rundir)
            aggregate_results_to_csv(output_dir, notebook_path)
        else:
            logger.error(f"{notebook_path} doesn't exist.")

    else:
        original_copy_ipynb = os.path.join(output_dir, f"{notebook_fname}.ipynb")
        shutil.copy(notebook_path, original_copy_ipynb)
        # Generate assertions for notebooks

        # 1. Find metrics and generate assertions

        property_finder = PropertyFinder(notebook_path, output_dir, original_nb_folder)
        property_finder.run()

        # 2. Run notebook to collect values used in assertions
        runner = NotebookRunner(
            notebook_path=tmp_ipynb,
            iterations=iterations,
            output_dir=output_dir
        )
        runner.run()

        # 3. Generate assertions based on the properties and values collected

        generator = AssertionGenerator(
            instrumentation_pkl=instrumentation_pkl,
            original_copy_ipynb=original_copy_ipynb,
            properties_csv=properties_csv,
            output_dir=output_dir,
            notebook_fname=notebook_fname,
            stats_method=stats_method,
            conf_level = conf_level,
            iteration = iterations
        )
        generator.run()

if __name__ == "__main__":
    sys.exit(main())



import ast
import os
import json
import sys
from collections import defaultdict
from pathlib import Path
import nbformat as nbf
import pandas as pd
import argparse

proj_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(proj_folder)
global config_path
config_path = os.path.join(proj_folder, "config.json")

from utils.utils import setup_logger

SHAP_PLOTS = ['bar', 'beeswarm', 'benchmark', 'decision', 'embedding', 'group_difference', 'heatmap',
    'image', 'monitoring', 'partial_dependence', 'scatter', 'violin', 'waterfall'
]
SHAP_PLOTS += [f"{item}_plot" for item in SHAP_PLOTS]

class FunctionTransformer(ast.NodeTransformer):

    def __init__(self, notebook_path, output_dir, output_notebook):
        self.output_file = output_notebook
        self.found_properties = []
        self.cell_no = 0
        self.notebook_path = notebook_path
        self.notebook_fname = Path(output_notebook).stem.replace("_tmp", "").split(".ipynb")[0]
        self.def_use_map = defaultdict(list)
        self.output_dir = output_dir
        os.makedirs(output_dir, exist_ok=True)
        self.def_set = set()
        self.need_instrument = False
        self.type_keywords = {
                "DATASET": ["ss", "ts", "ds", "train", "test_data", "dataset", "data", "correlation", "x_test", "df", "X_test", "y_val",
                             "X_train", "y_train", "X_val", "y_val"],
                "MODEL_PER": ["accuracy", "loss", "precision", "recall", "f1", "auc", "correct", "incorrect", "results", "error", "y_pred",
                              "acc"],
                "MODEL_ARCH": ["model", "rfr"]
            }


        with open(config_path, "r") as config_file:
            self.known_apis = json.load(config_file)["KNOWN_APIS"]

        logs_dir = os.path.join(output_dir, f"{self.notebook_fname}_logs")
        os.makedirs(logs_dir, exist_ok=True)
        self.logger = setup_logger(os.path.join(logs_dir, f'{self.__class__.__name__}.log'), self.__class__.__name__)
        self.logger_no_instrument = setup_logger(os.path.join(logs_dir, 'duplicate_assertions.log'), 'duplicate_assertions')
        self.tmpvar_count = 0
        self.tf_metrics = []

    def process_each_cell(self):
        ntbk = nbf.read(self.notebook_path, nbf.NO_CONVERT)

        for i, cell in enumerate(ntbk.cells):

            if cell.cell_type == "code":
                try:
                    code = cell.source
                    self.cell_no = i+1

                    tree = ast.parse(code)

                    transformed_tree = self.visit(tree)
                    cell.source = ast.unparse(transformed_tree)
                except Exception as e:
                    self.logger.error(f"Error processing cell {i+1}: {e}")

        nbf.write(ntbk, self.output_file)

    def add_property_dict(self, var_name, assert_type, assert_var, func_name, line):
        """
        Add a property dictionary to the list of found properties.
        """
        self.logger.info(f"Checking if ({var_name}, {self.cell_no},{line}) should be added.")
        self.logger.info("Current self.def_set: ")
        self.logger.info(self.def_set)

        def_key = (var_name, self.cell_no, line)

        def_value = f"{var_name}_{self.def_use_map.get(def_key)}"

        self.logger.info(f"The def_value is:")
        self.logger.info(def_value)
        if def_value not in self.def_set:
            self.logger.info(f"({var_name}, {self.cell_no},{line}) is not in the self.def_set. Instrumenting")
            # Only instrument the value with different def
            if "None" not in def_value:
                self.def_set.add(def_value)
            self.need_instrument = True

            property_dict = {
                var_name: {
                    'assert_type': f"{assert_type}__{func_name}",
                    'assert_var': assert_var,
                    'func_name': func_name,
                    'line': line,
                    'cell_no':self.cell_no
                }
            }
            self.found_properties.append(property_dict)
        else:
            self.need_instrument = False
            self.logger.info(f"({var_name}, {self.cell_no},{line}) is already in the self.def_set. Stop instrumenting")
            self.logger_no_instrument.info(f"({var_name}, {self.cell_no},{line})")

    def get_func_name(self, func):
        if isinstance(func, ast.Name):
            return func.id
        elif isinstance(func, ast.Attribute):
            return func.attr
        return None

    def visit_Assign(self, node):
        """
        Process assignment statements to identify relevant metrics.
        """

        if not isinstance(node.targets[0], (ast.Name, ast.Tuple)) or not isinstance(node.value, ast.Call):
            return node

        func_name = self.get_func_name(node.value.func)
        if func_name not in self.known_apis:
            return node

        new_nodes = []

        if func_name == "classification_report":
            assert_type = self.known_apis[func_name]["assert_type"]
            self.logger.info("classification_report")
            arg = node.value
            report_call = ast.Call(
                            func=ast.Name(id='classification_report', ctx = ast.Load()),
                            args = [ast.copy_location(a, arg) for a in arg.args],
                            keywords=[ast.copy_location(ast.keyword(
                                                arg=kw.arg,
                                                value = ast.copy_location(kw.value, kw.value)
                                                ), kw) for kw in arg.keywords]
                        )

            has_output_dict = False
            for keyword in report_call.keywords:
                if keyword.arg == 'output_dict':
                    has_output_dict = True
                    break

            if not has_output_dict:
                report_call.keywords.append(ast.keyword(arg='output_dict', value = ast.Constant(value=True)))

            value = ast.unparse(report_call)

            metric_list = [
                ("accuracy",),
                ("macro avg", "precision"),
                ("macro avg", "recall"),
                ("macro avg", "f1-score"),
                ("macro avg", "support"),
                ("weighted avg", "precision"),
                ("weighted avg", "recall"),
                ("weighted avg", "f1-score"),
                ("weighted avg", "support"),
            ]
            for metric_keys in metric_list:
                metric = ''.join(f'["{key}"]' for key in metric_keys)
                instrument_value =value+metric
                self.add_property_dict(instrument_value, assert_type, instrument_value, func_name, node.lineno)

                if self.need_instrument and assert_type != "UNKNOWN":
                    assert_node = self.create_instrument_node(instrument_value, self.cell_no, node.lineno, api=func_name)
                    insert_index = self.get_insert_position(node)
                    new_nodes.insert(insert_index, assert_node)

        elif func_name == "fit":
            arg_list = []
            for kw in node.value.keywords:
                arg_list.append(kw.arg)

            if "epochs" in arg_list or "validation_data" in arg_list or "batch_size" in arg_list:
                new_nodes = [node]
                value = ast.unparse(node.targets[0])

                metric_list = ["loss"]

                metric_list.extend(self.tf_metrics)

                if "validation_data" in arg_list:
                    val_list = []
                    for m in metric_list:
                        val_list.append("val_" + m)

                    metric_list.extend(val_list)

                for metric in metric_list:
                    instrument_value = value + f".history['{metric}'][-1]"
                    self.add_property_dict(instrument_value, "MODEL_PERF", instrument_value, func_name, node.lineno)

                    if self.need_instrument:
                        assert_node = self.create_instrument_node(instrument_value, self.cell_no, node.lineno, api=func_name)
                        insert_index = self.get_insert_position(node)
                        new_nodes.insert(insert_index, assert_node)

        elif func_name == "DataFrame":
            arg_list = []
            for arg in node.value.args:
                if isinstance(arg, ast.Attribute) and (arg.attr == "history"):
                    return node

            if isinstance(node.targets[0], ast.Name):
                target = node.targets[0]
                self.add_property_dict(target.id, "DATASET", target.id, func_name, node.lineno)
                assert_type = self.known_apis[func_name]["assert_type"]

                if self.need_instrument:
                    assert_node = self.create_instrument_node(target.id, self.cell_no, node.lineno, api=func_name)
                    insert_index = self.get_insert_position(node)
                    new_nodes.insert(insert_index, node)
                    new_nodes.insert(insert_index, assert_node)

        elif (func_name in SHAP_PLOTS):
            temp_var_name = f"nbtest_tmpvar_{self.tmpvar_count}"

            fig_assign = ast.Assign(
                targets=[ast.Name(id=temp_var_name, ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id="plt", ctx=ast.Load()),
                        attr="gcf",
                        ctx=ast.Load()
                    ),
                    args=[],
                    keywords=[]
                )
            )
            ast.fix_missing_locations(fig_assign)

            self.add_property_dict(
                var_name=temp_var_name,
                assert_type='PLOT',
                assert_var=temp_var_name,
                func_name=func_name,
                line=node.lineno
            )

            assert_node = self.create_instrument_node(f'nbtest_tmpvar_{self.tmpvar_count}', self.cell_no, node.lineno)
            self.tmpvar_count += 1

            return [node, fig_assign, assert_node]

        else:
            targets = node.targets[0].elts if isinstance(node.targets[0], ast.Tuple) else [node.targets[0]]
            new_nodes = [node]

            for target in targets:
                if isinstance(target, ast.Name):
                    var_name = target.id
                    assert_type = self.known_apis[func_name]["assert_type"]
                    assert_var = var_name

                    self.add_property_dict(var_name, assert_type, assert_var, func_name, node.lineno)

                    if self.need_instrument and assert_type != "UNKNOWN":
                        assert_node = self.create_instrument_node(var_name, self.cell_no, node.lineno, api=func_name)
                        insert_index = self.get_insert_position(node)
                        new_nodes.insert(insert_index, assert_node)

        if new_nodes:
            return new_nodes
        else:
            return [node]


    def visit_Expr(self, node):
        """
            Process expressions for implicit prints or outputs.
            e.g.,
            train_df.head()
            plt.show()
            "Test string".format(var)
            print(test_values.shape)
            print(X_test.shape)
        """
        new_nodes = [node]
        assert_type = "UNKNOWN"
        if isinstance(node.value, ast.Name):
            var_name = self.get_base_id(node.value)
            stored_value = var_name.lower()
            assert_type = self.determine_type(stored_value)

            self.add_property_dict(
                var_name=var_name,
                assert_type=assert_type,
                assert_var=var_name,
                func_name="SINGLE_VAR",
                line=node.lineno
            )

            if self.need_instrument and assert_type != "UNKNOWN":
                assert_node = self.create_instrument_node(var_name, self.cell_no, node.lineno)
                insert_index = self.get_insert_position(node)
                new_nodes.insert(insert_index, assert_node)
            return new_nodes

        # Handle attribute access (e.g., `train_data.shape`)
        if isinstance(node.value, ast.Attribute):
            var_name = ast.unparse(node.value)
            assert_type = self.determine_type(node.value)

            self.add_property_dict(
                var_name=var_name,
                assert_type=assert_type,
                assert_var=var_name,
                func_name="ATTRIBUTE_ACCESS",
                line=node.lineno
            )

            if self.need_instrument and assert_type != "UNKNOWN":
                assert_node = self.create_instrument_node(var_name, self.cell_no, node.lineno)
                new_nodes.append(assert_node)
            return new_nodes

        if isinstance(node.value, ast.Call):
            func_name = self.get_func_name(node.value.func)

            if func_name == "compile":
                self.tf_metrics = []
                for kw in node.value.keywords:
                    if kw.arg == "metrics" and isinstance(kw.value, ast.List):
                        for val in kw.value.elts:
                            if isinstance(val, ast.Constant) and isinstance(val.value, str):
                                self.tf_metrics.append(val.value)

            # Handle `print` function calls
            if func_name == "print":
                key = None
                assert_type = "PRINT"
                if node.value.args:
                    # If the first argument is a string literal, use it as the key
                    if isinstance(node.value.args[0], ast.Constant) and isinstance(node.value.args[0].value, str):
                        stored_value = self.get_base_id(node.value.args[0]).lower()
                        assert_type = self.determine_type(stored_value)
                        key = repr(node.value.args[0].value)

                    for arg in node.value.args:
                        # Handle nested function calls or variables
                        if isinstance(arg, ast.Call):
                            if isinstance(arg.func, ast.Attribute) and arg.func.attr == "format":
                                for format_arg in arg.args:  # Extract only arguments to .format()
                                    # Check for variable name
                                    assert_type = self.determine_type(node.value)

                                    extracted_value = ast.unparse(format_arg)
                                    self.add_property_dict(key or extracted_value, assert_type, extracted_value, "print", node.lineno)

                                    if self.need_instrument and assert_type != "UNKNOWN":
                                        assert_node = self.create_instrument_node(extracted_value, self.cell_no, node.lineno)
                                        insert_index = self.get_insert_position(node)
                                        new_nodes.insert(insert_index, assert_node)
                            else:
                                value = ast.unparse(arg)

                                # # Check for variable name
                                assert_type = self.determine_type(node.value)

                                func_name = self.get_func_name(arg.func)
                                if func_name in self.known_apis:
                                    assert_type = self.known_apis[func_name]["assert_type"]
                                else:
                                    func_name = "print_call"

                                if func_name == "classification_report":
                                    report_call = ast.Call(
                                                    func=ast.Name(id='classification_report', ctx = ast.Load()),
                                                    args = [ast.copy_location(a, arg) for a in arg.args],
                                                    keywords=[ast.copy_location(ast.keyword(
                                                                        arg=kw.arg,
                                                                        value = ast.copy_location(kw.value, kw.value)
                                                                        ), kw) for kw in arg.keywords]
                                                )

                                    has_output_dict = False
                                    for keyword in report_call.keywords:
                                        if keyword.arg == 'output_dict':
                                            has_output_dict = True
                                            break

                                    if not has_output_dict:
                                        report_call.keywords.append(ast.keyword(arg='output_dict', value = ast.Constant(value=True)))

                                    value = ast.unparse(report_call)

                                    metric_list = [
                                        ("accuracy",),
                                        ("macro avg", "precision"),
                                        ("macro avg", "recall"),
                                        ("macro avg", "f1-score"),
                                        ("macro avg", "support"),
                                        ("weighted avg", "precision"),
                                        ("weighted avg", "recall"),
                                        ("weighted avg", "f1-score"),
                                        ("weighted avg", "support"),
                                    ]
                                    for metric_keys in metric_list:
                                        metric = ''.join(f'["{key}"]' for key in metric_keys)
                                        instrument_value =value+metric
                                        self.add_property_dict(instrument_value, assert_type, instrument_value, func_name, node.lineno)

                                        if self.need_instrument and assert_type != "UNKNOWN":
                                            assert_node = self.create_instrument_node(instrument_value, self.cell_no, node.lineno, api=func_name)
                                            insert_index = self.get_insert_position(node)
                                            new_nodes.insert(insert_index, assert_node)

                                else:
                                    self.add_property_dict(key or value, assert_type, value, func_name, node.lineno)

                                    if self.need_instrument and assert_type != "UNKNOWN":
                                        assert_node = self.create_instrument_node(value, self.cell_no, node.lineno,api=func_name)
                                        insert_index = self.get_insert_position(node)
                                        new_nodes.insert(insert_index, assert_node)

                        elif isinstance(arg, ast.BinOp): #print('Train Accuracy:', l1.score(x_train, y_train) * 100)
                            if isinstance(arg.left, ast.Constant):
                                stored_value = self.get_base_id(arg.left).lower()
                                assert_type = self.determine_type(stored_value)

                            if isinstance(arg.right, ast.Call):
                                var_name = ast.unparse(arg.right)

                                self.add_property_dict(
                                    var_name=var_name,
                                    assert_type=assert_type,
                                    assert_var=var_name,
                                    func_name=func_name,
                                    line=node.lineno
                                )
                                if self.need_instrument and assert_type != "UNKNOWN":
                                    assert_node = self.create_instrument_node(var_name, self.cell_no, node.lineno)
                                    insert_index = self.get_insert_position(node)
                                    new_nodes.insert(insert_index, assert_node)

                            elif isinstance(arg.right, ast.Tuple):
                                for var in arg.right.elts:
                                    var_name = ast.unparse(var)

                                    self.add_property_dict(
                                        var_name=var_name,
                                        assert_type=assert_type,
                                        assert_var=var_name,
                                        func_name=func_name,
                                        line=node.lineno
                                    )
                                    if self.need_instrument and assert_type != "UNKNOWN":
                                        assert_node = self.create_instrument_node(var_name, self.cell_no, node.lineno)
                                        insert_index = self.get_insert_position(node)
                                        new_nodes.insert(insert_index, assert_node)


                        elif isinstance(arg, ast.Attribute):  # Handles cases like `test_values.shape`
                            var_name = ast.unparse(arg)
                            self.add_property_dict(
                                var_name=var_name,
                                assert_type=assert_type,
                                assert_var=var_name,
                                func_name=func_name,
                                line=node.lineno
                            )

                            if self.need_instrument and assert_type != "UNKNOWN":
                                assert_node = self.create_instrument_node(var_name, self.cell_no, node.lineno)
                                insert_index = self.get_insert_position(node)
                                new_nodes.insert(insert_index, assert_node)

                        elif isinstance(arg, ast.Name):  # Handle variables like `test_values`
                            var_name = arg.id
                            self.add_property_dict(
                                var_name=var_name,
                                assert_type=assert_type,
                                assert_var=var_name,
                                func_name="print",
                                line=node.lineno
                            )

                            if self.need_instrument and assert_type != "UNKNOWN":
                                assert_node = self.create_instrument_node(var_name, self.cell_no, node.lineno)
                                insert_index = self.get_insert_position(node)
                                new_nodes.insert(insert_index, assert_node)

                        # Handle f-strings
                        elif isinstance(arg, ast.JoinedStr):

                            for value in arg.values:
                                if isinstance(value, ast.Constant):
                                    stored_value = self.get_base_id(value).lower()
                                    assert_type = self.determine_type(stored_value)

                                if isinstance(value, ast.FormattedValue):
                                    extracted_value = ast.unparse(value.value)
                                    self.add_property_dict(key or extracted_value,assert_type, extracted_value, "print", node.lineno)

                                    if self.need_instrument and assert_type != "UNKNOWN":
                                        assert_node = self.create_instrument_node(extracted_value, self.cell_no, node.lineno)
                                        insert_index = self.get_insert_position(node)
                                        new_nodes.insert(insert_index, assert_node)

            # General case: Handle non-`print` function calls
            else:
                if func_name in self.known_apis:
                    assert_type = self.known_apis[func_name]["assert_type"]
                else:
                    assert_type = self.determine_type(node.value)

                # Handle .format() calls
                if isinstance(node.value.func, ast.Attribute) and node.value.func.attr == "format":
                    for arg in node.value.args:
                        if isinstance(arg, ast.Name):
                            extracted_value = arg.id
                        else:
                            extracted_value = ast.unparse(arg)

                        self.add_property_dict(
                            var_name=ast.unparse(node.value),
                            assert_type=assert_type,
                            assert_var=extracted_value,
                            func_name=node.value.func.attr,
                            line=node.lineno
                        )

                        if self.need_instrument and assert_type != "UNKNOWN":
                            assert_node = self.create_instrument_node(extracted_value, self.cell_no, node.lineno)
                            insert_index = self.get_insert_position(node)
                            new_nodes.insert(insert_index, assert_node)

                elif isinstance(node.value.func, ast.Attribute) and node.value.func.attr == "fit":
                    arg_list = []
                    for kw in node.value.keywords:
                        arg_list.append(kw.arg)

                    if "epochs" in arg_list or "validation_data" in arg_list or "batch_size" in arg_list:
                        temp_var_name = f"nbtest_tmpvar_{self.tmpvar_count}"

                        # Create an assignment node for the temporary variable
                        assign_node = ast.Assign(
                            targets=[ast.Name(id=temp_var_name, ctx=ast.Store())],
                            value=node.value
                        )
                        ast.fix_missing_locations(assign_node)

                        metric_list = ["loss"]

                        metric_list.extend(self.tf_metrics)

                        if "validation_data" in arg_list:
                            val_list = []
                            for m in metric_list:
                                val_list.append("val_" + m)

                            metric_list.extend(val_list)

                        insert_index = self.get_insert_position(node)

                        res = []
                        res.append(assign_node)

                        for metric in metric_list:
                            instrument_value = temp_var_name + f".history['{metric}'][-1]"
                            self.add_property_dict(instrument_value, "MODEL_PERF", instrument_value, "fit", node.lineno)

                            assert_node = self.create_instrument_node(instrument_value, self.cell_no, node.lineno, api=func_name)
                            res.append(assert_node)

                        return res

                elif isinstance(node.value.func, ast.Attribute) and func_name == "show" and ast.unparse(node) == "plt.show()":
                    temp_var_name = f"nbtest_tmpvar_{self.tmpvar_count}"

                    fig_assign = ast.Assign(
                        targets=[ast.Name(id=temp_var_name, ctx=ast.Store())],
                        value=ast.Call(
                            func=ast.Attribute(
                                value=ast.Name(id="plt", ctx=ast.Load()),
                                attr="gcf",
                                ctx=ast.Load()
                            ),
                            args=[],
                            keywords=[]
                        )
                    )
                    ast.fix_missing_locations(fig_assign)

                    self.add_property_dict(
                        var_name=temp_var_name,
                        assert_type=assert_type,
                        assert_var=temp_var_name,
                        func_name=func_name,
                        line=node.lineno
                    )

                    assert_node = self.create_instrument_node(temp_var_name, self.cell_no, node.lineno)
                    insert_index = self.get_insert_position(node)
                    self.tmpvar_count += 1

                    return [fig_assign, assert_node, node]

                elif isinstance(node.value.func, ast.Attribute) and (func_name in SHAP_PLOTS) and (ast.unparse(node).startswith('shap.')):
                    show_present, show_kw = False, ast.keyword(arg='temp', value=ast.Constant(None))
                    for kw in node.value.keywords:
                        if (kw.arg == 'show'):
                            show_present = True
                            show_kw = kw
                            break;

                    if show_present:
                        show_kw.value = ast.Constant(value=False)
                    else:
                        node.value.keywords.append(ast.keyword(arg='show', value=ast.Name(id='False', ctx=ast.Load())))

                    temp_var_name = f"nbtest_tmpvar_{self.tmpvar_count}"

                    fig_assign = ast.Assign(
                        targets=[ast.Name(id=temp_var_name, ctx=ast.Store())],
                        value=ast.Call(
                            func=ast.Attribute(
                                value=ast.Name(id="plt", ctx=ast.Load()),
                                attr="gcf",
                                ctx=ast.Load()
                            ),
                            args=[],
                            keywords=[]
                        )
                    )
                    ast.fix_missing_locations(fig_assign)

                    self.add_property_dict(
                        var_name=temp_var_name,
                        assert_type=assert_type,
                        assert_var=temp_var_name,
                        func_name=func_name,
                        line=node.lineno
                    )

                    assert_node = self.create_instrument_node(f'nbtest_tmpvar_{self.tmpvar_count}', self.cell_no, node.lineno)
                    insert_index = self.get_insert_position(node)
                    self.tmpvar_count += 1

                    return [node, fig_assign, assert_node]

                else:
                    # General function call handling
                    temp_var_name = f"nbtest_tmpvar_{self.tmpvar_count}"

                    # Create an assignment node for the temporary variable
                    assign_node = ast.Assign(
                        targets=[ast.Name(id=temp_var_name, ctx=ast.Store())],
                        value=node.value
                    )
                    ast.fix_missing_locations(assign_node)

                    if self.need_instrument and assert_type != "UNKNOWN":
                        # Add property and assertion for the temporary variable
                        self.add_property_dict(
                            var_name=temp_var_name,
                            assert_type=assert_type,
                            assert_var=temp_var_name,
                            func_name=func_name,
                            line=node.lineno
                        )
                        assert_node = self.create_instrument_node(temp_var_name, self.cell_no, node.lineno, api=func_name)
                        self.tmpvar_count += 1
                        # Replace the original node with the assign_node and append the assert_node
                        parent = getattr(node, "parent", None)
                        if parent and hasattr(parent, "body"):
                            body = parent.body
                            index = body.index(node)
                            body[index] = assign_node
                            body.insert(index + 1, assert_node)

                        return [assign_node, assert_node]

        return new_nodes

    def create_instrument_node(self, var_name, cell_no, line_no, api=None):

        instrument_call = ast.Expr(
            value=ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id='nbtest', ctx=ast.Load()),
                    attr='instrument',
                    ctx=ast.Load()
                ),
                args=[
                    ast.Name(id=var_name, ctx=ast.Load()),
                    ast.Constant(value=cell_no, kind=None),
                    ast.Constant(value=line_no, kind=None),
                    ast.Constant(value=self.notebook_fname, kind=None),
                    ast.Constant(value=var_name, kind=None)
                ],
                keywords=[ast.keyword(arg='api', value=ast.Constant(value=api))] if api else []
            )
        )
        ast.fix_missing_locations(instrument_call)
        return instrument_call

    def get_insert_position(self, node):
        """
        Determine the position to insert the assert_node based on the current and next line.
        """
        parent = getattr(node, "parent", None)
        if not parent or not hasattr(parent, "body"):
            return 1

        body = parent.body
        current_index = body.index(node)

        # If this is the last node, append at the end
        if current_index == len(body) - 1:
            return len(body)

        # Otherwise, insert before the next node
        return current_index + 1

    def get_base_id(self,node):
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Call):
            return self.get_base_id(node.func)
        elif isinstance(node, ast.Attribute):
            return self.get_base_id(node.value)
        elif isinstance(node, ast.Constant):
            return node.value
        else:
            return None

    def determine_type(self, node_value):
        stored_value = self.get_base_id(node_value)
        if stored_value is not None:
            stored_value = stored_value.lower()
            lowered = stored_value.lower()
            for type_name, keywords in self.type_keywords.items():
                if any(keyword in lowered for keyword in keywords):
                    return type_name
        return "UNKNOWN"

def main():

    code = """
print(classification_report(y_val, y_pred))
"""
    parser = argparse.ArgumentParser()
    parser.add_argument('notebook', type=str, help="notebook to parse")
    parser.add_argument('--output_dir', type=str, help="output folder")
    parser.add_argument('--output_notebook', type =str, help="output notebook path")

    args = parser.parse_args()

    function_transformer = FunctionTransformer(args.notebook, args.output_dir, args.output_notebook)
    function_transformer.process_each_cell()

    for row in function_transformer.found_properties:
        print(row)

if __name__ == "__main__":
    main()



import os
import sys
import json
import argparse
import nbformat as nbf
import ast
import pandas as pd
from pathlib import Path

proj_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(proj_folder)

from properties.functionTransformer import FunctionTransformer
from utils.utils import setup_logger, preprocess_code, set_seed_random_state

config_path = os.path.join(proj_folder, "config.json")
with open(config_path, "r") as config_file:
    CONFIG_FILEPATH = json.load(config_file)["FILE_PATH"]

class PropertyFinder:
    def __init__(self, notebook, output_dir, original_nb_dir):
        self.notebook = notebook
        self.output_dir = output_dir
        self.notebook_fname = Path(notebook).stem
        self.original_nb_dir = original_nb_dir
        self.tmp_ipynb = os.path.join(self.original_nb_dir, f"{self.notebook_fname}_{CONFIG_FILEPATH['TMP_IPYNB']}")
        self.processed_ipynb = os.path.join(self.output_dir, f"{self.notebook_fname}_{CONFIG_FILEPATH['PROCESSED_IPYNB']}")
        self.properties_csv = os.path.join(output_dir, f"{self.notebook_fname}_{CONFIG_FILEPATH['PROPERTIES_CSV']}")
        self.properties_df = pd.DataFrame(columns=["Variable", "Assertion Type", "Code Cell", "Line Number"])

        # Setup logging
        logs_dir = os.path.join(output_dir, f"{self.notebook_fname}_logs")
        os.makedirs(logs_dir, exist_ok=True)
        self.logger = setup_logger(os.path.join(logs_dir, f'{self.__class__.__name__}.log'), self.__class__.__name__)

    def find_property(self):
        """
        Analyze notebook cells and extract properties using FunctionTransformer.
        """
        try:
            ntbk = nbf.read(self.notebook, as_version=4)
        except:
            self.logger.error(f"File not exist: {self.notebook_fname}")
            return

        nb_dest = nbf.v4.new_notebook()
        # Insert import cell
        import_cell = nbf.v4.new_code_cell(source="import nbtest\nimport json\nimport numpy as np\nimport matplotlib.pyplot as plt\nrandom_seed = np.random.randint(10000)")
        nb_dest.cells.append(import_cell)

        # Preprocess code
        for i, cell in enumerate(ntbk.cells):
            if cell.cell_type == "code":
                code = cell.source
                # Ignore HTML, JavaScript, or magic commands
                if not any(code.lstrip().startswith(prefix) for prefix in ["%%html", "%%javascript", "%%bash", "%%sh", "%%latex"]):
                    cleaned_code_no_seed = preprocess_code(cell.source)
                    cleaned_code = set_seed_random_state(cleaned_code_no_seed)
                    cleaned_cell = nbf.v4.new_code_cell(cleaned_code)
                    nb_dest.cells.append(cleaned_cell)
            else:
                nb_dest.cells.append(cell)

        exit_cell = nbf.v4.new_code_cell(source="exit()")
        nb_dest.cells.append(exit_cell)

        nbf.write(nb_dest, self.processed_ipynb)

        function_transformer = FunctionTransformer(self.processed_ipynb, self.output_dir, self.tmp_ipynb)
        function_transformer.process_each_cell()

        for property_dict in function_transformer.found_properties:
            for var_name, info in property_dict.items():
                temp_df = pd.DataFrame({
                    "Variable": [info["assert_var"]],
                    "Assertion Type": [info["assert_type"]],
                    "Code Cell": [info["cell_no"]],
                    "Line Number": [info["line"]],
                })
                self.properties_df = pd.concat([self.properties_df, temp_df], ignore_index=True)

        # nbf.write(ntbk, self.tmp_ipynb)
        self.properties_df.to_csv(self.properties_csv, index=False)
        self.logger.info(f"Properties saved to {self.properties_csv}")

    def run(self):
        """
        Run the entire property finding and notebook updating process.
        """
        self.find_property()

def main():
    parser = argparse.ArgumentParser(description="Generate properties for a Jupyter notebook.")
    parser.add_argument("notebook", type=str, help="Path to the Jupyter notebook file")
    parser.add_argument("--output_dir", required=True, help="Directory to save output files")
    parser.add_argument("--original_dir", help="Directory to original folders with notebook")
    args = parser.parse_args()

    notebook_path = os.path.abspath(args.notebook)
    output_dir = os.path.abspath(args.output_dir)
    original_dir = os.path.abspath(args.original_dir)

    property_finder = PropertyFinder(notebook_path, output_dir, original_dir)
    property_finder.run()

if __name__ == "__main__":
    main()



import argparse
import os
import glob
import pickle
import subprocess
import sys
import json
from pathlib import Path

proj_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(proj_folder)
from utils.utils import setup_logger

config_path = os.path.join(proj_folder, "config.json")
global CONFIG_FILEPATH
with open(config_path, "r") as config_file:
    CONFIG_FILEPATH = json.load(config_file)["FILE_PATH"]

class NotebookRunner:
    def __init__(self, notebook_path, iterations, output_dir):
        self.notebook_path = os.path.abspath(notebook_path)
        self.iterations = iterations
        self.output_dir = os.path.abspath(output_dir)
        self.prev_ite = []
        self.notebook_fname = Path(self.notebook_path).stem.replace("_tmp", "").split(".ipynb")[0]

        self.instrumentation_pkl = os.path.join(output_dir, f"{self.notebook_fname}_{CONFIG_FILEPATH['INSTRUMENTATION_PKL']}")

        self.logs_dir = os.path.join(output_dir, f"{self.notebook_fname}_logs")
        os.makedirs(self.logs_dir, exist_ok=True)
        self.logger = setup_logger(os.path.join(self.logs_dir, f'{self.__class__.__name__}.log'), self.__class__.__name__)

    def parse_previous_iterations(self):
        pattern = f"{self.notebook_fname}_instrumentation_*.pkl"
        for file in sorted(glob.glob(os.path.join(self.output_dir, pattern))):
            num = int(file.split("_")[-1].replace(".pkl", ""))
            self.prev_ite.append(num)

    def execute_notebook(self, input_path, output_path):
        try:
            result = subprocess.run(
                [
                    "jupyter",
                    "nbconvert",
                    "--to",
                    "notebook",
                    "--execute",
                    input_path,
                    "--output",
                    output_path
                ],
                capture_output=True,
                text=True,
                check=True,
                env={
                    **os.environ,
                    "COLLECT_VARS": "1",
                    "NBTEST_OUTPUT_DIR": self.output_dir
                }
            )
            self.logger.debug(f"Execution output: {result.stdout}")
            self.logger.debug(f"Execution output: {result.stderr}")
        except Exception as e:

            self.logger.error(f"Error when executing {self.notebook_path}", exc_info=True)

            self.logger.debug(f"Execution errors: {e.stderr}")

    def merge_pickle_files(self, output_file):
        ''' Read all pkl files with format: instrumentation_{num}.pkl
            Merge these pkl files into output_file (instrumentation.pkl)
            {num} is the pkl file for the ith iteration. {num} starts from 1
        '''
        self.logger.info(f"self.notebook_fname: {self.notebook_fname}")
        pickle_files = glob.glob(os.path.join(self.output_dir, f"{self.notebook_fname}_instrumentation_*.pkl"))
        if not pickle_files:
            self.logger.warning(f"No instrumentation files found: {self.notebook_fname}_instrumentation_*.pkl")
            return

        loaded_lists = []

        for file in pickle_files:
            with open(file, "rb") as f:
                loaded_lists.append(pickle.load(f))

        merged_list = [{} if isinstance(item, dict) else None for item in loaded_lists[0]]

        for lst in loaded_lists:
            for i, item in enumerate(lst):
                if (len(item) == 1 and next(iter(item.values()))["isNone"] is True) or merged_list[i] is None:
                    continue

                elif isinstance(item, dict):
                    for key, value in item.items():
                        if key not in merged_list[i]:
                            merged_list[i][key] = {}
                        for sub_key, sub_value in value.items():
                            if (sub_key == "cell_no" or sub_key == "line_no" or sub_key == "isNone"):
                                merged_list[i][key][sub_key] = sub_value
                            else:
                                if sub_key not in merged_list[i][key]:
                                    merged_list[i][key][sub_key] = []
                                merged_list[i][key][sub_key].append(sub_value)

        with open(output_file, "wb") as f:
            pickle.dump(merged_list, f)


    def run(self):
        self.parse_previous_iterations()

        for i in range(self.iterations):
            output_path = os.path.join(self.output_dir, f"{self.notebook_fname}_run_{i}.ipynb")
            if (i + 1) not in self.prev_ite:
                self.logger.debug(f"------------------ Iteration {i + 1} ---------------------------")
                self.execute_notebook(self.notebook_path, output_path)

        self.merge_pickle_files(self.instrumentation_pkl)

def main():
    parser = argparse.ArgumentParser(description="Run a notebook multiple times.")
    parser.add_argument("notebook", type=str, help="Path to the notebook file.")
    parser.add_argument("-n", type=int, default=1, help="Number of times to run the notebook.")
    parser.add_argument("-o", default=os.getcwd(), help="Output directory")


    args = parser.parse_args()

    runner = NotebookRunner(
        notebook_path=args.notebook,
        iterations=args.n,
        output_dir=args.o
    )
    runner.run()


if __name__ == "__main__":
    main()



import logging
import re
import nbformat as nbf
from io import StringIO
from pathlib import Path

def setup_logger(log_file_path, logger_name, level=logging.DEBUG):

    logger = logging.getLogger(logger_name)
    logger.setLevel(level)

    file_handler = logging.FileHandler(log_file_path)
    file_handler.setLevel(level)

    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)

    logger.propagate = False

    return logger

def set_seed_random_state(code):
    # Step 4: Replace 'random_state' argument values with 'random_seed'

    random_state_pattern = r"(random_state\s*=\s*)\d+(?=\s*[),])"
    cleaned_code = re.sub(random_state_pattern, r"\1random_seed", code)

    return cleaned_code

def preprocess_code(code):
    """
    Preprocess the code by:
    1. Removing lines starting with '!', '%', or containing 'pip install'.
    2. Removing redundant os.walk blocks that print file paths.
    3. Removing all comments and blank lines.
    """
    # Step 1: Remove lines with '!', '%', or 'pip install'
    cleaned_lines = [
        line for line in code.splitlines()
        if not (line.strip().startswith('!') or line.strip().startswith('%') or 'pip install' in line)
    ]
    cleaned_code = "\n".join(cleaned_lines)

    # Step 2: Remove os.walk blocks
    os_walk_pattern = (
        r"for\s+[\w\d_]+,\s*_[^:]*in\s+os\.walk\([^)]+\):\s*\n"
        r"(?:\s*#.*\n)*"  # Optional comments
        r"\s*for\s+[\w\d_]+[^:]*:\s*\n"
        r"(?:\s*#.*\n)*"  # Optional comments
        r"\s*print\(.*\)\s*"
    )
    cleaned_code = re.sub(os_walk_pattern, '', cleaned_code, flags=re.MULTILINE)

    # Step 3: Remove comments and blank lines using tokenize
    final_lines = []
    for line in cleaned_code.splitlines():
        stripped_line = line.strip()
        # Exclude blank lines and comments, preserve other lines with original spaces
        if stripped_line and not stripped_line.startswith("#"):
            final_lines.append(line)
    cleaned_code = "\n".join(final_lines)

    return cleaned_code

def get_notebook_name(notebook_path):
    notebook_stat_name = Path(notebook_path).stem
    notebook_fname = notebook_stat_name.split("_chebyshev")[0]
    return notebook_fname

def main():
    ntbk = nbf.read("/home/yy2282/project/nb_test/results/results/MAIN__chebyshev_2_0.95/kaggle__brokerus__ml-project-house-prices-eda-and-7-models/ml-project-house-prices-eda-and-7-models.ipynb", nbf.NO_CONVERT)
    for i, cell in enumerate(ntbk.cells):
        if cell.cell_type == "code":
            cleaned_code = preprocess_code(cell.source)
            print(cleaned_code)

if __name__ == "__main__":
    main()



import pandas as pd
import argparse
import pickle

parser = argparse.ArgumentParser(description="Read and print a pickle file.")
parser.add_argument('file', type=str, help='Path to the pickle file')
args = parser.parse_args()


with open(args.file, "rb") as f:
    data = pickle.load(f)
    for item in data:
        print(item)



import os
import glob
import re
import subprocess
import pandas as pd
import logging
import sys
from pathlib import Path
import shutil
from pathlib import Path

proj_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(proj_folder)

from utils.utils import setup_logger


# Collect iterations already run
def get_completed_iterations(output_dir, notebook_fname):

    pattern = os.path.join(output_dir, f'{notebook_fname}_pytest*.txt')
    return [
        int(match.group(1)) for file_path in glob.glob(pattern)
        if (match := re.search(r'pytest(\d+)\.txt', file_path))  # Match .txt files
    ]


# Run missing tests
def run_missing_tests(output_dir, notebook_path, pytest_iterations, run_folder):
    notebook_stat_name = Path(notebook_path).stem
    notebook_fname = notebook_stat_name.split("_chebyshev")[0]

    is_copy = False

    if run_folder is None:
        run_folder = os.path.dirname(notebook_path)
    else:
        if os.path.dirname(notebook_path) != run_folder:
            original_copy_ipynb = os.path.join(run_folder, f"{notebook_stat_name}.ipynb")
            if os.path.abspath(notebook_path) != os.path.abspath(original_copy_ipynb):

                shutil.copy(notebook_path, original_copy_ipynb)
                # if os.path.exists(original_copy_ipynb):
                #     print(f"Copying {notebook_path} to {original_copy_ipynb}")
                is_copy = True


            notebook_path = original_copy_ipynb

    prev_ite = get_completed_iterations(output_dir, notebook_fname)

    for i in range(1, pytest_iterations + 1):
        if i not in prev_ite:
            iter_txt_path = os.path.join(output_dir, f'{notebook_fname}_pytest{i}.txt')

            logs_dir = os.path.join(output_dir, f"{notebook_fname}_logs")
            os.makedirs(logs_dir, exist_ok=True)
            log_file_path = os.path.join(logs_dir, f'{notebook_fname}_pytest{i}.log')

            logger = setup_logger(log_file_path, "iteration_{i}")

            logger.info(f"Starting iteration {i}...")

            logger.info(f"Running pytest on {notebook_path}")

            result = subprocess.run(
                ["pytest", "--nbtest", "-v", notebook_path,
                "--nbtest-output-dir", output_dir,
                "--nbtest-log-filename", os.path.join(output_dir, f"{notebook_fname}_pytest_testid_res_{i}.csv"),
                "--nbtest-nblog-name", os.path.join(output_dir, f"{notebook_fname}_pytest_nb_tracebacks_{i}.ipynb")],
                capture_output=True,
                text=True,
                cwd=run_folder
            )

            logger.info(f"STDOUT:\n{result.stdout}")
            logger.debug(f"STDERR:\n{result.stderr}")

            status = "Fail" if result.returncode not in (0, 5) else "Pass"

            with open(iter_txt_path, "w") as file:
                file.write(status)
    if is_copy:
        Path(notebook_path).unlink() # Delete file


# Aggregate results into a CSV file
def aggregate_results_to_csv(output_dir, notebook_path):

    notebook_stat_name = Path(notebook_path).stem
    notebook_fname = notebook_stat_name.split("_chebyshev")[0]

    completed_iterations = get_completed_iterations(output_dir, notebook_fname)

    rows = []
    total_pass = 0
    total_fail = 0
    for iteration in completed_iterations:
        iter_txt_path = os.path.join(output_dir, f'{notebook_fname}_pytest{iteration}.txt')
        try:
            with open(iter_txt_path, "r") as file:
                status = file.read().strip()
                rows.append({
                    "Iteration": iteration,
                    "Pass": 1 if status == "Pass" else 0,
                    "Fail": 1 if status == "Fail" else 0
                })

                if status == "Pass":
                    total_pass +=1
                if status == "Fail":
                    total_fail +=1
        except FileNotFoundError:
            logging.error(f"File not found: {iter_txt_path}")

    if not rows:
        logging.debug("No data found to aggregate.")
        return

    df = pd.DataFrame(rows).sort_values(by="Iteration").reset_index(drop=True)
    output_csv_path = os.path.join(output_dir, f'{notebook_fname}_summary.csv')
    df.to_csv(output_csv_path, index=False)

    output_cnt_path = os.path.join(output_dir, f'{notebook_fname}_total_cnt.csv')
    df = pd.DataFrame({'Total_run': [total_pass+total_fail], 'Total_pass':[total_pass], 'Total_fail':[total_fail]})
    df.to_csv(output_cnt_path, index=False)

def main():
    # Example usage
    output_dir = "/home/yy2282/project/nb_test/testing-jupyter-notebook/outputs"
    notebook_path = "/home/yy2282/project/nb_test/testing-jupyter-notebook/outputs/predicting-alternate-dimension-transportation_chebyshev_0.95_1.ipynb"
    pytest_iterations = 1
    run_dir = "/home/yy2282/project/nb_test/testing-jupyter-notebook/examples/kaggle__abdelazizsami__predicting-alternate-dimension-transportation"


    # Run tests and aggregate results
    run_missing_tests(output_dir, notebook_path, pytest_iterations, run_dir)
    # aggregate_results_to_csv(output_dir,notebook_path)

if __name__ == "__main__":
    main()



import ast
import os
import sys
import nbformat as nbf
import re
import pickle
from pathlib import Path

proj_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(proj_folder)
from utils.utils import setup_logger, preprocess_code

import ast

SHAP_PLOTS = ['bar', 'beeswarm', 'benchmark', 'decision', 'embedding', 'group_difference', 'heatmap',
    'image', 'monitoring', 'partial_dependence', 'scatter', 'violin', 'waterfall'
]
SHAP_PLOTS += [f"{item}_plot" for item in SHAP_PLOTS]

class FunctionTransformer(ast.NodeTransformer):
    def __init__(self, input_nb, logs_dir, output_nb, mapping):

        self.input_nb = input_nb
        self.output_nb = output_nb
        self.mapping = mapping

        self.assertion_list = []
        self.modified_lines = set()
        self.assertion_generated = {"Assertion_id": [], "Assertion": [], "Assertion_type": []}

        self.output_dir = Path(output_nb).parent.absolute()

        self.test_id = 0
        self.cell_no = 0
        self.logger = setup_logger(os.path.join(logs_dir, f'assertion_{self.__class__.__name__}.log'), self.__class__.__name__)

        self.curr_cell_metadata = []

    def process_each_cell(self):
        ntbk = nbf.read(self.input_nb, nbf.NO_CONVERT)

        for cell_dict in self.mapping:
            for cell_no, assertions_config in cell_dict.items():
                self.logger.info(f"No. Cell: {cell_no}. Retriving: ntbk.cells[{cell_no-1}] - {assertions_config}")
                self.assertion_list = assertions_config
                cell = ntbk.cells[cell_no-1]

                self.cell_no = cell_no - 1

                code = cell.source
                self.logger.info(code)
                tree = ast.parse(code)
                transformed_tree = self.visit(tree)

                cell.source = ast.unparse(transformed_tree)
                cell.metadata["nbtest_hidden_asserts"] = self.curr_cell_metadata
                self.curr_cell_metadata = []

        nbf.write(ntbk, self.output_nb, version=4)

    def visit(self, node):
        prev_var = set()
        num_asserts = 0
        for idx, stmt in enumerate(node.body):

            for assertion in self.assertion_list:
                self.logger.info("Checking if this assertion should be added.")
                self.logger.info(assertion)

                if not self.is_assert_node(stmt):
                    if hasattr(stmt, 'lineno') and stmt.lineno == (assertion["lineno"]):
                        var_name = assertion["var"]

                        if var_name not in prev_var and var_name.startswith("nbtest_tmpvar"):
                            if assertion["func_name"] == "assert_plot_equal":
                                call = ast.Expr(ast.Call(
                                    func=ast.Attribute(
                                        value=ast.Name(id="plt", ctx=ast.Load()),
                                        attr="gcf",
                                        ctx=ast.Load()
                                    ),
                                    args=[],
                                    keywords=[]
                                ))
                                ast.fix_missing_locations(call)

                                assign_node = self.create_assignment_node(var_name, call)
                                if ast.unparse(stmt) == "plt.show()":
                                    node.body.insert(idx, assign_node)
                                elif ('shap.' in ast.unparse(stmt)) and (any([(i in ast.unparse(stmt) for i in SHAP_PLOTS)])):
                                    show_present, show_kw = False, ast.keyword(arg='temp', value=ast.Constant(None))

                                    for kw in stmt.value.keywords:
                                        if (kw.arg == 'show'):
                                            show_present = True
                                            show_kw = kw
                                            break

                                    if show_present:
                                        show_kw.value = ast.Constant(value=False)
                                    else:
                                        stmt.value.keywords.append(ast.keyword(arg='show', value=ast.Name(id='False', ctx=ast.Load())))

                                    node.body[idx] = stmt

                                    node.body.insert(idx+1, assign_node)
                            else:
                                assign_var_name = var_name
                                if ".history[" in var_name:
                                    assign_var_name = var_name.split(".history")[0]

                                if assign_var_name not in prev_var:
                                    assign_node = self.create_assignment_node(assign_var_name, stmt)
                                    node.body[idx] = assign_node
                                    prev_var.add(assign_var_name)

                        pattern_shape = r"(?P<var>\w+)\.shape\b"
                        match_shape = re.match(pattern_shape, assertion["args"][0])
                        if match_shape:
                            assertion["func_name"] = "assert_shape"
                            assertion["args"][0] = match_shape.group("var")

                        pattern_dtype_listcomp = (
                            r"\[\s*str\s*\(\s*(?P<var>\w+)\s*\[\s*i\s*\]\s*\.s*dtype\s*\)"
                            r"\s*for\s+i\s+in\s+sorted\s*\(\s*(?P=var)\s*\.s*columns\s*\)\s*\]"
                        )
                        match_column_types = re.match(pattern_dtype_listcomp, assertion["args"][0])
                        if match_column_types:
                            assertion["func_name"] = "assert_column_types"
                            assertion["args"][0] = match_column_types.group("var")

                        match_column_names = re.match(r"sorted\s*\(\s*(?P<var>\w+)\s*\.s*columns\s*\)", assertion["args"][0])
                        if match_column_names:
                            assertion["func_name"] = "assert_column_names"
                            assertion["args"][0] = match_column_names.group("var")

                        pattern_nanmean_expr = (
                            r"np\.nanmean\s*\(\s*(?P<var>\w+)\.select_dtypes\s*\(\s*include\s*=\s*\[\s*['\"]number['\"]\s*\]\s*\)"
                            r"\.to_numpy\s*\(\s*\)\s*\)"
                        )
                        match_nanmean = re.match(pattern_nanmean_expr, assertion["args"][0])
                        if match_nanmean:
                            assertion["func_name"] = "assert_df_mean"
                            assertion["args"][0] = match_nanmean.group("var")

                        pattern_nanvar_expr = (
                            r"np\.nanvar\s*\(\s*(?P<var>\w+)\.select_dtypes\s*\(\s*include\s*=\s*\[\s*['\"]number['\"]\s*\]\s*\)"
                            r"\.to_numpy\s*\(\s*\)\s*\)"
                        )
                        match_nanvar = re.match(pattern_nanvar_expr, assertion["args"][0])
                        if match_nanvar:
                            assertion["func_name"] = "assert_df_var"
                            assertion["args"][0] = match_nanvar.group("var")

                        pattern_sklearn_params = (
                            r"\{\s*k\s*:\s*v\s+for\s+k\s*,\s*v\s+in\s+(?P<model>\w+)\.get_params\(\)\.items\(\)\s+"
                            r"if\s+k\s*!=\s*['\"]random_state['\"]\s*"
                            r"and\s+not\s*\(\s*hasattr\s*\(\s*v\s*,\s*['\"]__module__['\"]\s*\)\s*"
                            r"and\s*v\.__module__\.startswith\(\s*['\"]sklearn['\"]\s*\)\s*\)\s*\}"
                        )
                        match_sklearn_model = re.match(pattern_sklearn_params, assertion["args"][0])
                        if match_sklearn_model:
                            model_fname = os.path.join(self.output_dir, f"sklearn_model_{self.test_id}.pkl")
                            assertion["func_name"] = "assert_sklearn_model"
                            model_params = assertion["args"][1]
                            with open(model_fname, 'wb') as f:
                                pickle.dump(model_params, f)

                            assertion["args"] = [var_name, os.path.relpath(model_fname, start=os.path.dirname(os.getcwd()))]

                        pattern_layers = r"\[\s*\(\s*layer\.__class__\.__name__,\s*layer\.output_shape,\s*layer\.count_params\(\)\s*\)\s*for\s+\w+\s+in\s+(?P<variable>\w+)\.layers\s*\]"
                        match_json_loads = re.match(pattern_layers, assertion["args"][0])
                        if match_json_loads:
                            assertion["func_name"] = "assert_nn_model"
                            assertion["args"][0] = var_name

                        self.logger.info(assertion["func_name"])
                        self.logger.info(assertion["args"])

                        assert_node = self.create_assert_node(
                            stmt.lineno+1,
                            stmt.col_offset,
                            assertion["func_name"],
                            assertion["args"],
                            assertion["kwargs"],
                            f'{self.test_id}'
                        )

                        if assert_node is not None:
                            if var_name in prev_var and assertion["func_name"] == "assert_plot_equal":
                                continue
                            assert_idx = 0
                            if (idx + 1 < len(node.body)) and re.match(r"^nbtest_tmpvar_\d+\s*=\s*plt\.gcf\(\)\s*$", ast.unparse(node.body[idx + 1])):
                                assert_node.lineno += 1
                                # node.body.insert(idx + 2, assert_node)
                                assert_idx = idx + 2
                            else:
                                # node.body.insert(idx + 1, assert_node)
                                assert_idx = idx + 1

                            self.curr_cell_metadata.append({"index": assert_idx + num_asserts, "content": ast.unparse(assert_node)})

                            num_asserts += 1

                            prev_var.add(var_name)
                            args_part = ", ".join(str(arg) for arg in assertion["args"])

                            if assertion["kwargs"]:
                                kwargs_part = ", ".join([f"{key}={value}" for key, value in assertion["kwargs"].items()])
                                full_assert = f'nbtest.{assertion["func_name"]}({args_part}, {kwargs_part})'
                            else:
                                full_assert = f'nbtest.{assertion["func_name"]}({args_part})'

                            self.assertion_generated["Assertion_id"].append(f'{self.test_id}')
                            self.assertion_generated["Assertion"].append(full_assert)
                            self.assertion_generated["Assertion_type"].append(assertion["assert_type"])

                            self.test_id += 1

        return node

    def create_assignment_node(self, var_name, stmt):
        """
        Create an assignment node when the variable name starts with "nbtest_tmpvar".
        """
        return ast.Assign(
            targets=[ast.Name(id=var_name, ctx=ast.Store())],
            value=stmt.value,
            lineno=stmt.lineno,
            col_offset=stmt.col_offset
        )

    def create_assert_node(self, lineno, col_offset, func_name, args, kwargs, test_id):
        """
        Create an assertion node dynamically based on provided function name and arguments.
        """
        args = args if args is not None else []
        kwargs = kwargs if kwargs is not None else {}


        ast_args = []
        if args:
            ast_args.append(ast.Name(id=args[0], ctx=ast.Load()))
            # ast_args.extend(self.to_ast_node(arg) for arg in args[1:])

            for arg in args[1:]:
                ast_arg = self.to_ast_node(arg)
                if ast_arg is None:
                    return None

                ast_args.append(ast_arg)


            ast_keywords = []
            for key, value in kwargs.items():
                ast_node = self.to_ast_node(value)
                if ast_node is None:
                    return None
                ast_keywords.append(ast.keyword(arg=key, value=ast_node))

            ast_keywords.append(ast.keyword(arg='test_id', value=self.to_ast_node(test_id)))

            return ast.Expr(
                value=ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(id='nbtest', ctx=ast.Load()),
                        attr=func_name,
                        ctx=ast.Load()
                    ),
                    args=ast_args,
                    keywords=ast_keywords
                ),
                lineno=lineno,
                col_offset=col_offset
            )


    def to_ast_node(self, value):
        """Convert a value to an AST node."""
        if isinstance(value, ast.AST):
            return value
        elif isinstance(value, (int, float)):
            return ast.Constant(value=value)
        elif isinstance(value, str):
            return ast.Constant(value=value)
        elif isinstance(value, list):
            return ast.List(elts=[self.to_ast_node(v) for v in value], ctx=ast.Load())
        elif isinstance(value, tuple):
            return ast.Tuple(elts=[self.to_ast_node(v) for v in value], ctx=ast.Load())
        elif isinstance(value, dict):
            return ast.Dict(
                keys=[self.to_ast_node(k) for k in value.keys()],
                values=[self.to_ast_node(v) for v in value.values()]
            )
        elif value is None:
            return ast.Constant(value=None)
        else:
            self.logger.warning(f"Unsupported type of {value}: {type(value)}")
            return ast.Constant(value=None)

    def is_assert_node(self, stmt):
        """Check if stmt is an assertion node created by `create_assert_node`."""
        return (
            isinstance(stmt, ast.Expr) and
            isinstance(stmt.value, ast.Call) and
            isinstance(stmt.value.func, ast.Attribute) and
            isinstance(stmt.value.func.value, ast.Name) and
            stmt.value.func.value.id == "nbtest"  # Ensures it's calling nbtest.<func_name>
        )


def main():
    source_code = """
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
from tensorflow.keras.optimizers import Adam
Sequential([Dense(x_train.shape[1], activation='relu'), Dense(24, activation='relu'), Dense(12, activation='relu'), Dense(10, activation='relu'), Dense(1, activation='sigmoid')])
model.compile(loss='binary_crossentropy', optimizer=Adam(0.03), metrics=['Accuracy'])
history = model.fit(x_train, y_train, epochs=50, validation_batch_size=64, validation_data=(x_val, y_val))
    """

    cleaned_code = preprocess_code(source_code)


    assertion_list = [
        {
            "var": "nbtest_tmpvar_1",
            "lineno": 4,
            "func_name": "assert_allclose",
            "args": ["nbtest_tmpvar_1", "expected_value"],
            "kwargs": {"atol": 0.0003632869200276443},
            "assert_type":"model"
        },
        {
            "var": "model",
            "lineno": 5,
            "func_name": "assert_allclose",
            "args": ["model", "expected_value"],
            "kwargs": {"atol": 0.0003632869200276443},
            "assert_type":"model"
        },
        {
            "var": "test",
            "lineno": 4,
            "func_name": "assert_allclose",
            "args": ["test", "expected_value"],
            "kwargs": {"atol": 0.0003632869200276443},
            "assert_type":"model"
        }
    ]


    tree = ast.parse(cleaned_code)

    transformer = FunctionTransformer(assertion_list)
    modified_tree = transformer.visit(tree)

    modified_code = ast.unparse(modified_tree)
    print("\nModified Source Code:\n", modified_code)


if __name__ == "__main__":
    main()



import pandas as pd
import nbformat as nbf
import os
import numpy as np
import sys
import pickle
import traceback
from collections import defaultdict
import json
import shutil
from matplotlib.testing.compare import compare_images
import re

proj_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

sys.path.append(proj_folder)
from utils.utils import setup_logger, preprocess_code
from assert_gen.functionTransformer import FunctionTransformer

config_path = os.path.join(proj_folder, "config.json")
with open(config_path, "r") as config_file:
    CONFIG_FILEPATH = json.load(config_file)["FILE_PATH"]

def normalize_dict(d):
    return {
        k: v.tolist() if isinstance(v, np.ndarray) else v
        for k, v in d.items()
    }

class AssertionGenerator:
    def __init__(self, instrumentation_pkl, original_copy_ipynb,properties_csv, output_dir, notebook_fname, stats_method, conf_level, iteration):
        """
        Initialize the AssertionGenerator with required file paths and parameters.
        """
        self.instrumentation_pkl = instrumentation_pkl
        self.original_copy_ipynb = original_copy_ipynb
        self.properties_csv = properties_csv
        self.output_dir = output_dir
        self.notebook_fname = notebook_fname
        self.stats_method = stats_method
        self.conf_level = conf_level
        self.iteration = iteration
        self.processed_ipynb = os.path.join(self.output_dir, f"{self.notebook_fname}_{CONFIG_FILEPATH['PROCESSED_IPYNB']}")

        self.mapping = {}

        self.logs_dir = os.path.join(output_dir, f"{self.notebook_fname}_logs")
        os.makedirs(self.logs_dir, exist_ok=True)
        self.logger = setup_logger(os.path.join(self.logs_dir, f'{self.__class__.__name__}.log'), self.__class__.__name__)

        try:
            self.instrument_value = pd.read_pickle(self.instrumentation_pkl)
        except Exception as e:
            self.logger.error(f"Failed to read pickle file {self.instrumentation_pkl}: {e}")
            raise

        try:
            self.properties_df = pd.read_csv(self.properties_csv)
        except Exception as e:
            self.logger.error(f"Failed to read CSV file {self.properties_csv}: {e}")
            raise

        try:
            self.ntbk = nbf.read(self.original_copy_ipynb, nbf.NO_CONVERT)
        except Exception as e:
            self.logger.error(f"Failed to read notebook file {self.original_copy_ipynb}: {e}")
            raise

    def calculate_stats_bound(self, values, method, confidence_level=0.95):
        """
        Calculate the relative tolerance and mean value based on statistical methods.
        """
        try:

            values = [float(x) for x in values]

            mean_val = np.mean(values)
            std_dev = np.std(values, ddof=0)

            if method == "chebyshev":
                epsilon = np.sqrt(1 / (1 - confidence_level)) * std_dev
            elif method == "hoeffding":
                epsilon = np.sqrt(np.log(2 / (1 - confidence_level)) / (2 * len(values)))
        except Exception as e:
            self.logger.error(f"Error in calcualte bound for {values}" + traceback.format_exc())
            return None

        return mean_val, epsilon


    def get_assert_type(self, value):

        assert_type = "assert_equal"

        if all(isinstance(item, (int, float, np.number)) for item in value):
            assert_type = "assert_allclose"

        if all(isinstance(item, bool) for item in value):
            assert_type = f"assert_{str(value[0]).lower()}"

        if not isinstance(value, list) or len(value) == 0:
            return "unsupported"

        if isinstance(value, list) or isinstance(value, pd.Series):
            first_value = value.iloc[0] if isinstance(value, pd.Series) else value[0]
            if first_value is None:
                return "unsupported"

        if all(isinstance(item, dict) for item in value):
            if isinstance(value, (pd.Series, np.ndarray)):
                value = value.tolist()
            first_dict = normalize_dict(value[0])
            if not all(normalize_dict(d) == first_dict for d in value):
                return "unsupported"

        if isinstance(value[0], (list, tuple)):
            if len(value[0]) == 0:
                return "unsupported"
            if not isinstance(value[0][0], (int, float, str, bool, tuple)):
                return "unsupported"

        if isinstance(value[0], str) and value[0].startswith(str(self.output_dir)):
            assert_type = "assert_plot_equal"

        return assert_type


    def decide_assert_type_bound(self, assertion_config):
        instrument_dict = assertion_config["val"]
        assert_var = assertion_config["var"]
        assertion_config["func_name"] = None
        assertion_config["args"] = []
        assertion_config["kwargs"] = {}

        if instrument_dict:
            for feature, value in instrument_dict.items():
                assert_type = self.get_assert_type(value)
                updated_feature = feature.replace("assert_var", assert_var)

                if assert_type == "assert_allclose":
                    results = self.calculate_stats_bound(
                        value,
                        self.stats_method,
                        self.conf_level
                    )
                    if results is not None:
                        assert_values, assert_rtol = results
                        assertion_config["func_name"] = assert_type
                        assertion_config["args"] = [str(updated_feature), assert_values]
                        assertion_config["kwargs"] = {"atol": assert_rtol}
                        continue
                elif assert_type == "unsupported":
                    continue
                elif assert_type == "assert_plot_equal":
                    try:
                        # Use the first image as reference
                        ref_img_path = value[0]
                        rms_errors = []

                        for path in value[1:]:
                            result = compare_images(ref_img_path, path, tol=0)
                            if result is None:
                                rms_errors.append(0.0)
                            else:
                                # Use regex to extract RMS value from error string
                                match = re.search(r"RMS Value:\s*([0-9.]+)", result)
                                if match:
                                    rms_value = float(match.group(1))
                                    rms_errors.append(rms_value)

                        rms_threshold = self.calculate_stats_bound(
                            rms_errors,
                            self.stats_method,
                            self.conf_level
                        )

                        # Persist the reference image to a known path
                        persistent_ref_img = os.path.join(self.output_dir, f"reference_plot_{assert_var}.png")
                        shutil.copy(ref_img_path, persistent_ref_img)

                        assertion_config["func_name"] = assert_type
                        assertion_config["args"] = [assert_var, os.path.basename(self.output_dir) + f"/reference_plot_{assert_var}.png"]
                        assertion_config["kwargs"] = {"tol": rms_threshold[0]}

                        self.logger.info(f"Assertion config: {str(assertion_config)}")

                    except Exception as e:
                        self.logger.info(f"Error occurred while generating plot assertion:\n{e}")
                        continue

                else:
                    first_item = value[0]
                    if isinstance(first_item, dict):
                        first_item = normalize_dict(first_item)
                        has_same_item = all((first_item == normalize_dict(item)) for item in value[1:])
                    else:
                        has_same_item = all((first_item == item) for item in value[1:])

                    if has_same_item:
                        assertion_config["func_name"] = assert_type
                        assertion_config["args"] = [str(updated_feature), value[0]]
                        assertion_config["kwargs"] = {}
                    else:
                        continue

        return assertion_config

    def generate_assertions(self):

        assertion_generated = {"Assertion_id": [], "Assertion": [], "Assertion_type": []}

        ntbk = nbf.read(self.original_copy_ipynb, as_version=4)

        nb_dest = nbf.v4.new_notebook()

        # Insert import cell
        import_cell = nbf.v4.new_code_cell(source="import nbtest\nimport json\nimport numpy as np\nimport matplotlib.pyplot as plt\nrandom_seed = np.random.randint(10000)")
        nb_dest.cells.append(import_cell)

        # Preprocess code
        for i, cell in enumerate(ntbk.cells):
            if cell.cell_type == "code":
                code = cell.source
                # Ignore HTML, JavaScript, or magic commands
                if not any(code.lstrip().startswith(prefix) for prefix in ["%%html", "%%javascript", "%%bash", "%%sh", "%%latex"]):
                    cleaned_code_no_seed = preprocess_code(cell.source)
                    cleaned_cell = nbf.v4.new_code_cell(cleaned_code_no_seed)
                    nb_dest.cells.append(cleaned_cell)
            else:
                nb_dest.cells.append(cell)

        nbf.write(nb_dest, self.processed_ipynb)

        self.logger.info(f"In generate_assertions: {self.mapping}")

        for cell_dict in self.mapping:
            self.logger.info(cell_dict)
            for cell_no, assertions_config in cell_dict.items():
                for assertion_config in assertions_config:
                    self.decide_assert_type_bound(assertion_config)

                assertions_config[:] = [ac for ac in assertions_config if ac["func_name"] is not None]

        notebook_w_assert = os.path.join(self.output_dir, f"{self.notebook_fname}_{self.stats_method}_{self.conf_level}_{self.iteration}.ipynb")

        function_transformer = FunctionTransformer(self.processed_ipynb, self.logs_dir, notebook_w_assert, self.mapping)
        function_transformer.process_each_cell()

        assertion_generated = {key: assertion_generated[key] + function_transformer.assertion_generated[key] for key in assertion_generated}
        df = pd.DataFrame(assertion_generated)
        df.to_csv(os.path.join(self.output_dir, f"{self.notebook_fname}_assertions.csv"), index=False)

    def map_var_to_property(self):
        self.logger.info("Creating mapping between the property and the instrumentated value.")
        properties_df = pd.read_csv(self.properties_csv)

        with open(self.instrumentation_pkl, "rb") as f:
            pickle_data = pickle.load(f)

        cell_mapping = defaultdict(list)

        self.logger.info(f"Checking the contents in {self.instrumentation_pkl}")
        for pickle_row in pickle_data:
            self.logger.info(pickle_row)
            self.logger.info('\n')

        for _, row in properties_df.iterrows():
            variable = row["Variable"]
            assertion_type = row["Assertion Type"]
            code_cell = row["Code Cell"]
            line_number = row["Line Number"]
            is_assertion = False

            for pickle_row in pickle_data:
                if pickle_row:
                    variable_name = next(iter(pickle_row.keys()))
                    instrument_info = next(iter(pickle_row.values()))

                    if variable_name == variable:

                        if instrument_info["cell_no"] == code_cell and instrument_info["line_no"] == line_number:
                            values = {key: value for key, value in instrument_info.items() if key not in {"cell_no", "line_no", "isNone"}}

                            for feature, value in values.items():

                                assertion_entry = {
                                    "var": variable,
                                    "assert_type": assertion_type,
                                    "lineno": line_number,
                                    "val": {feature: value}
                                }

                                cell_mapping[code_cell].append(assertion_entry)
                                is_assertion = True

            if is_assertion:
                self.logger.info(f"{variable} at Cell {code_cell} Line {line_number} has found the instrumentation value")
            else:
                self.logger.info(f"{variable} at Cell {code_cell} Line {line_number} didn't find the instrumentation value. No matching code_cell and line_number or variable name.")


        self.mapping = [{cell_no: assertions} for cell_no, assertions in cell_mapping.items()]
        self.logger.info("Checking self.mapping")
        for mapping in self.mapping:
            self.logger.info(mapping)

    def run(self):
        if os.path.isfile(self.instrumentation_pkl):
            self.map_var_to_property()
            self.generate_assertions()
        else:
            self.logger.warning(f"No assertions generated because {self.instrumentation_pkl} is not found. ")


# Example Usage
if __name__ == "__main__":
    instrumentation_pkl = "/home/yy2282/project/nb_test/testing-jupyter-notebook/outputs/house-prices-advanced-regression-techniques_instrumentation.pkl"
    original_copy_ipynb = "/home/yy2282/project/nb_test/testing-jupyter-notebook/outputs/house-prices-advanced-regression-techniques.ipynb"
    properties_csv = "/home/yy2282/project/nb_test/testing-jupyter-notebook/outputs/house-prices-advanced-regression-techniques_properties.csv"
    output_dir = "/home/yy2282/project/nb_test/testing-jupyter-notebook/outputs"
    notebook_fname = "house-prices-advanced-regression-techniques"
    stats_method = "chebyshev"
    conf_level = 0.95
    iteration = 2

    processor = AssertionGenerator(
        instrumentation_pkl=instrumentation_pkl,
        original_copy_ipynb = original_copy_ipynb,
        properties_csv=properties_csv,
        output_dir=output_dir,
        notebook_fname=notebook_fname,
        stats_method=stats_method,
        conf_level = conf_level,
        iteration = iteration
    )

    processor.run()
